@@ -1,4 +1,4 @@
-/* Copyright (C) 2019 Artifex Software, Inc.
+/* Copyright (C) 2019-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -112,19 +112,9 @@ static const pl_interp_characteristics_t png_characteristics = {
     png_detect_language,
     "Artifex",
     "0.01",
-    "11 Nov 2019",
-    1 /* minimum input size */
+    "11 Nov 2019"
 };
 
-/* GS's fakakta png integration insists on putting a
- * memory structure pointer in the decompress structs client data.
- * This is no good to find our instance from. We therefore find
- * it by offsetting back from the address of the cinfo. This is
- * a nasty bit of casting, so wrap it in a macro. */
-#define JPG_FROM_CINFO(CINFO) \
-    (png_interp_instance_t *)(((char *)(CINFO))-offsetof(png_interp_instance_t,cinfo))
-
-
 /* Get implementation's characteristics */
 static const pl_interp_characteristics_t * /* always returns a descriptor */
 png_impl_characteristics(const pl_interp_implementation_t *impl)     /* implementation of interpreter to alloc */
@@ -210,39 +200,6 @@ png_impl_get_device_memory(pl_interp_implementation_t *impl)
     return png->dev ? png->dev->memory : NULL;
 }
 
-#if 0 /* UNUSED */
-static int
-png_impl_set_param(pl_interp_implementation_t *impl,
-                   pl_set_param_type           type,
-                   const char                 *param,
-                   const void                 *val)
-{
-    png_interp_instance_t *png = (png_interp_instance_t *)impl->interp_client_data;
-
-    /* No params set here */
-    return 0;
-}
-
-static int
-png_impl_add_path(pl_interp_implementation_t *impl,
-                  const char                 *path)
-{
-    png_interp_instance_t *png = (png_interp_instance_t *)impl->interp_client_data;
-
-    /* No paths to add */
-    return 0;
-}
-
-static int
-png_impl_post_args_init(pl_interp_implementation_t *impl)
-{
-    png_interp_instance_t *png = (png_interp_instance_t *)impl->interp_client_data;
-
-    /* No post args processing */
-    return 0;
-}
-#endif
-
 /* Prepare interp instance for the next "job" */
 static int
 png_impl_init_job(pl_interp_implementation_t *impl,
@@ -256,25 +213,6 @@ png_impl_init_job(pl_interp_implementation_t *impl,
     return 0;
 }
 
-#if 0 /* UNUSED */
-static int
-png_impl_run_prefix_commands(pl_interp_implementation_t *impl,
-                             const char                 *prefix)
-{
-    png_interp_instance_t *png = (png_interp_instance_t *)impl->interp_client_data;
-
-    return 0;
-}
-
-static int
-png_impl_process_file(pl_interp_implementation_t *impl, const char *filename)
-{
-    png_interp_instance_t *png = (png_interp_instance_t *)impl->interp_client_data;
-
-    return 0;
-}
-#endif
-
 /* Do any setup for parser per-cursor */
 static int                      /* ret 0 or +ve if ok, else -ve error code */
 png_impl_process_begin(pl_interp_implementation_t * impl)
@@ -389,6 +327,20 @@ my_png_malloc(png_structp png_ptr, png_alloc_size_t size)
 {
     png_interp_instance_t *png = (png_interp_instance_t *)png_get_mem_ptr(png_ptr);
 
+    if (sizeof(void *) == 8) {
+        /* gs_alloc_bytes returns blocks aligned to 8 on 64bit platforms.
+         * PNG (on Windows at least) requires blocks aligned to 16. */
+        unsigned char *block = gs_alloc_bytes(png->memory, size+16, "my_png_malloc");
+        intptr_t num_bytes_padded;
+
+        if (block == NULL)
+            return NULL;
+        num_bytes_padded = 16-(((intptr_t)block) & 15);
+        block += num_bytes_padded;
+        block[-1] = num_bytes_padded;
+
+        return block;
+    }
     return gs_alloc_bytes(png->memory, size, "my_png_malloc");
 }
 
@@ -397,6 +349,13 @@ my_png_free(png_structp png_ptr, png_voidp ptr)
 {
     png_interp_instance_t *png = (png_interp_instance_t *)png_get_mem_ptr(png_ptr);
 
+    if (sizeof(void *) == 8) {
+        unsigned char *block = ptr;
+        if (block == NULL)
+            return;
+        block -= block[-1];
+        ptr = (void *)block;
+    }
     gs_free_object(png->memory, ptr, "my_png_free");
 }
 
@@ -634,7 +593,9 @@ do_impl_process(png_interp_instance_t *png, stream_cursor_read * pr, bool eof)
                 break;
             }
 
-            png->samples = gs_alloc_bytes(png->memory, png->byte_width * (png->interlaced ? png->height : 1), "png_impl_process(samples)");
+            png->samples = gs_alloc_bytes(png->memory,
+                                          (size_t)png->byte_width * (png->interlaced ? png->height : 1),
+                                          "png_impl_process(samples)");
             if (png->samples == NULL) {
                 png->state = ii_state_flush;
                 break;
@@ -715,16 +676,11 @@ do_impl_process(png_interp_instance_t *png, stream_cursor_read * pr, bool eof)
                 break;
             }
 
-            /* FIXME */
             png->state = ii_state_flush;
             break;
         }
         default:
         case ii_state_flush:
-            //png_destroy_decompress(&png->cinfo);
-
-            //gs_png_mem_term((j_common_ptr)&png->cinfo);
-
             if (png->penum) {
                 (void)gs_image_cleanup_and_free_enum(png->penum, png->pgs);
                 png->penum = NULL;
