@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -90,7 +90,7 @@ check_token(int token_type, const char *s, const char *e, int *score)
 
 #define TOKEN_CHECK(n) else if (e-s == strlen(n) && memcmp(s, n, e-s) == 0) { score[0] += e-s; score[1]++; }
 
-    if (0) {}
+    if (0) { /* Initial block of checking chain */ }
     TOKEN_CHECK("dup")
     TOKEN_CHECK("exch")
     TOKEN_CHECK("grestore")
@@ -121,7 +121,7 @@ score_comment(const char *s, const char *e, int *score)
 {
 #define COMMENT_CHECK(n) else if (e-s >= strlen(n) && memcmp(s, n, strlen(n)) == 0) { score[0] += 100; score[1]++; }
 
-    if (0) {}
+    if (0) { /* Initial block of checking chain */ }
     COMMENT_CHECK("!PS")
     COMMENT_CHECK("%Title:")
     COMMENT_CHECK("%Version:")
@@ -222,7 +222,7 @@ ps_detect_language(const char *s, int len)
             } else if (*t == 27) {
                 /* Give up if we meet an ESC. It could be a UEL. */
                 break;
-            } else if (*t <= 32 || *t > 127) {
+            } else if (*t <= 32 || *(unsigned char *)t > 127) {
                 if (check_token(token_type, token, t, score))
                     break;
                 if (*t != 9 && *t != 10 && *t != 12 && *t != 13 && *t != 32)
@@ -309,8 +309,7 @@ ps_impl_characteristics(const pl_interp_implementation_t *impl)     /* implement
     ps_detect_language,
     "Artifex",
     PSVERSION,
-    PSBUILDDATE,
-    1				/* minimum input size to PostScript */
+    PSBUILDDATE
   };
 #undef PSVERSION
 #undef PSBUILDDATE
@@ -379,13 +378,11 @@ ps_impl_get_device_memory(pl_interp_implementation_t *impl)
 
 static int
 ps_impl_set_param(pl_interp_implementation_t *impl,
-                  pl_set_param_type           type,
-                  const char                 *param,
-                  const void                 *val)
+                  gs_param_list *plist)
 {
     ps_interp_instance_t *psi = (ps_interp_instance_t *)impl->interp_client_data;
 
-    return psapi_set_param(psi->psapi_instance, type, param, val);
+    return psapi_set_param(psi->psapi_instance, plist);
 }
 
 static int
@@ -417,11 +414,10 @@ ps_impl_init_job(pl_interp_implementation_t *impl,
 {
     ps_interp_instance_t *psi = (ps_interp_instance_t *)impl->interp_client_data;
     int exit_code;
-    int code = 0;
+    int code;
 
     /* Any error after here *must* reset the device to null */
-    if (code >= 0)
-        code = psapi_set_device(psi->psapi_instance, device);
+    code = psapi_set_device(psi->psapi_instance, device);
 
     if (code >= 0)
         code = psapi_run_string(psi->psapi_instance, "erasepage", 0, &exit_code);
@@ -483,9 +479,67 @@ static int
 ps_impl_process(pl_interp_implementation_t * impl, stream_cursor_read * pr)
 {
     ps_interp_instance_t *psi = (ps_interp_instance_t *)impl->interp_client_data;
-    const unsigned int len = pr->limit - pr->ptr;
+    unsigned int len;
     int code, exit_code = 0;
 
+    if (psi->bytes_fed == 0)
+    {
+        /* Skip over whitespace/comments looking for a PDF marker. */
+        while (pr->ptr < pr->limit)
+        {
+            int i;
+
+            /* Skip over whitespace (as defined in PLRM) */
+            if (pr->ptr[1] == 0 ||
+                pr->ptr[1] == 9 ||
+                pr->ptr[1] == 10 ||
+                pr->ptr[1] == 12 ||
+                pr->ptr[1] == 13 ||
+                pr->ptr[1] == 32) {
+                pr->ptr++;
+                continue;
+            }
+
+            /* If we're not starting a comment, exit. */
+            if (pr->ptr[1] != '%')
+                break;
+
+            /* If we're starting with a PDF header, swap to file mode. */
+            if (pr->limit - pr->ptr >= 8 &&
+                strncmp((const char *)&pr->ptr[2], "PDF-", 4) == 0 &&
+                (pr->ptr[6] >= '1' && pr->ptr[6] <= '9') &&
+                pr->ptr[7] == '.' &&
+                (pr->ptr[8] >= '0' && pr->ptr[8] <= '9'))
+                return_error(gs_error_NeedFile);
+
+            /* Check for a historical PDF header. */
+            if (pr->limit - pr->ptr >= 22 &&
+                strncmp((const char *)&pr->ptr[2], "!PS-Adobe-", 10) == 0 &&
+                (pr->ptr[12] >= '0' && pr->ptr[12] <= '9') &&
+                pr->ptr[13] == '.' &&
+                (pr->ptr[14] >= '0' && pr->ptr[14] <= '9') &&
+                strncmp((const char *)&pr->ptr[15], " PDF-", 5) == 0 &&
+                (pr->ptr[20] >= '0' && pr->ptr[20] <= '9') &&
+                pr->ptr[21] == '.' &&
+                (pr->ptr[22] >= '0' && pr->ptr[22] <= '9'))
+                return_error(gs_error_NeedFile);
+
+            /* Do we have a complete comment that we can skip? */
+            for (i = 1; pr->ptr + i < pr->limit; i++)
+                if (pr->ptr[i+1] == 10 || pr->ptr[i+1] == 13) {
+                    pr->ptr += i;
+                    i = 0; /* Loop again in case there are more comments. */
+                    break;
+                }
+            /* If we fall out of the loop naturally, then we hit the end
+             * of the buffer without terminating our comment. We need to
+             * abort the loop and return. */
+            if (i != 0)
+                return_error(gs_error_NeedInput);
+        }
+    }
+
+    len = pr->limit - pr->ptr;
     code = psapi_run_string_continue(psi->psapi_instance, (const char *)pr->ptr + 1, len, 0, &exit_code);
     if (exit_code == gs_error_InterpreterExit) {
         int64_t offset;
