@@ -1,4 +1,4 @@
-/* Copyright (C) 2019 Artifex Software, Inc.
+/* Copyright (C) 2019-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -105,8 +105,7 @@ static const pl_interp_characteristics_t pwg_characteristics = {
     pwg_detect_language,
     "Artifex",
     "0.01",
-    "13 Nov 2019",
-    1 /* minimum input size */
+    "13 Nov 2019"
 };
 
 /* Get implementation's characteristics */
