@@ -1,4 +1,4 @@
-/* Copyright (C) 2019 Artifex Software, Inc.
+/* Copyright (C) 2019-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -26,7 +26,9 @@
 #include "gspaint.h"
 #include "plmain.h"
 #include "tiffio.h"
+#if defined(SHARE_JPEG) && SHARE_JPEG==0
 #include "jmemcust.h"
+#endif
 #include "gsmchunk.h"
 
 /* Forward decls */
@@ -90,7 +92,9 @@ typedef struct tiff_interp_instance_s {
 
     byte              *samples;
     byte              *proc_samples;
+#if defined(SHARE_JPEG) && SHARE_JPEG==0
     jpeg_cust_mem_data jmem;
+#endif
 } tiff_interp_instance_t;
 
 static int
@@ -112,8 +116,7 @@ static const pl_interp_characteristics_t tiff_characteristics = {
     tiff_detect_language,
     "Artifex",
     "0.01",
-    "18 Nov 2019",
-    1 /* minimum input size */
+    "18 Nov 2019"
 };
 
 /* Get implementation's characteristics */
@@ -421,7 +424,7 @@ static toff_t tifsSizeProc(thandle_t tiff_)
     return tiff->buffer_full;
 }
 
-#if defined(SHARE_LIBTIFF) && SHARE_LIBTIFF==0
+#if defined(SHARE_JPEG) && SHARE_JPEG==0
 static void *gs_j_mem_alloc(j_common_ptr cinfo, size_t size)
 {
     gs_memory_t *mem = (gs_memory_t *)(GET_CUST_MEM_DATA(cinfo)->priv);
@@ -472,7 +475,7 @@ tiff_jpeg_mem_callback(thandle_t tiff_)
 
     return &tiff->jmem;
 }
-#endif /* SHARE_LIBTIFF == 0 */
+#endif /* SHARE_JPEG == 0 */
 
 static int
 guess_pal_depth(int n, uint16_t *rmap, uint16_t *gmap, uint16_t *bmap)
@@ -619,7 +622,7 @@ do_impl_process(pl_interp_implementation_t * impl, stream_cursor_read * pr, int
                 tiff->photometric == PHOTOMETRIC_LOGLUV) {
                 TIFFSetField(tiff->handle, TIFFTAG_SGILOGDATAFMT, SGILOGDATAFMT_8BIT);
             }
-#if defined(SHARE_LIBTIFF) && SHARE_LIBTIFF==0
+#if defined(SHARE_JPEG) && SHARE_JPEG==0
             TIFFSetJpegMemFunction(tiff->handle,
                                    &tiff_jpeg_mem_callback);
 #endif
@@ -684,7 +687,6 @@ do_impl_process(pl_interp_implementation_t * impl, stream_cursor_read * pr, int
             }
             tiff->proc_samples = tiff->samples;
 
-            /* FIXME: Need to allow for LAB, YUV etc too */
             tiff->bpp = tiff->bpc * tiff->num_comps;
             switch(tiff->photometric) {
             case PHOTOMETRIC_MINISWHITE:
@@ -747,9 +749,9 @@ do_impl_process(pl_interp_implementation_t * impl, stream_cursor_read * pr, int
                     code = gs_error_unknownerror;
                     goto fail_decode;
                 } else if (tiff->tiled) {
-                    tiff->proc_samples = gs_alloc_bytes(tiff->memory, tiff->tile_width * tiff->tile_height * 3, "tiff_tile");
+		  tiff->proc_samples = gs_alloc_bytes(tiff->memory, (size_t)tiff->tile_width * tiff->tile_height * 3, "tiff_tile");
                 } else {
-                    tiff->proc_samples = gs_alloc_bytes(tiff->memory, tiff->width * 3, "tiff_scan");
+		  tiff->proc_samples = gs_alloc_bytes(tiff->memory, (size_t)tiff->width * 3, "tiff_scan");
                 }
                 break;
             }
@@ -760,6 +762,8 @@ do_impl_process(pl_interp_implementation_t * impl, stream_cursor_read * pr, int
                 }
                 break;
             case PHOTOMETRIC_SEPARATED:
+                if (tiff->num_comps == 3 || tiff->num_comps == 4)
+                    break;
             case PHOTOMETRIC_YCBCR:
             case PHOTOMETRIC_CIELAB:
             case PHOTOMETRIC_ICCLAB:
