@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -179,7 +179,7 @@ pcl_secondary_pitch(pcl_args_t * pargs, pcl_state_t * pcs)
     return pcl_pitch(float_arg(pargs), pcs, 1);
 }
 
-/* 
+/*
  * Experiments on HP printers indicate height command parameters are truncated to
  * the documented range .25 - 999.75.
  */
@@ -199,7 +199,7 @@ static int
 pcl_height(pcl_args_t * pargs, pcl_state_t * pcs, int set)
 {
 
-    uint height_4ths = 
+    uint height_4ths =
         (uint) (height_clamp(float_arg(pargs)) * 4 + 0.5);
     pcl_font_selection_t *pfs = &pcs->font_selection[set];
     pfs->params.height_4ths = height_4ths;
@@ -675,13 +675,6 @@ pcl_unload_resident_fonts(pcl_state_t * pcs)
     }
 }
 
-/* Purge all */
-static bool
-purge_all(const gs_memory_t * mem, cached_char * cc, void *dummy)
-{
-    return true;
-}
-
 static int
 pcfont_do_reset(pcl_state_t * pcs, pcl_reset_type_t type)
 {
@@ -728,33 +721,8 @@ pcfont_do_reset(pcl_state_t * pcs, pcl_reset_type_t type)
         pl_dict_release(&pcs->simm_fonts);
 
         if (pcs->font_dir) {
-            gx_purge_selected_cached_chars(pcs->font_dir,
-                                           purge_all,
-                                           (void *)NULL);
-            /* free character cache machinery */
-            gs_free_object(pcs->font_dir->memory, pcs->font_dir->fmcache.mdata, "pcsfont_do_reset");
-            {
-                /* free the circular list of memory chunks first */
-                gx_bits_cache_chunk *chunk = pcs->font_dir->ccache.chunks;
-                gx_bits_cache_chunk *start_chunk = chunk;
-                gx_bits_cache_chunk *prev_chunk;
-                while (1) {
-                    if (start_chunk == chunk->next) {
-                        gs_free_object(pcs->font_dir->ccache.bits_memory, chunk->data, "pcsfont_do_reset");
-                        gs_free_object(pcs->font_dir->ccache.bits_memory, chunk, "pcsfont_do_reset");
-                        break;
-                    }
-                    prev_chunk = chunk;
-                    chunk = chunk->next;
-                    gs_free_object(pcs->font_dir->ccache.bits_memory, prev_chunk->data, "pcsfont_do_reset");
-                    gs_free_object(pcs->font_dir->ccache.bits_memory, prev_chunk, "pcsfont_do_reset");
-                }
-
-                pcs->font_dir->ccache.chunks = NULL;
-                gs_free_object(pcs->font_dir->memory, pcs->font_dir->ccache.table, "pcfont_do_reset");
-                gs_free_object(pcs->font_dir->memory, pcs->font_dir, "pcfont_do_reset");
-                pcs->font_dir = 0;
-            }
+            gs_free_object(pcs->font_dir->memory, pcs->font_dir, "pcfont_do_reset");
+            pcs->font_dir = 0;
         }
     }
     return 0;
